const {ServiceBroker} = require('moleculer')
const ConsoleLogger = require('moleculer').Loggers.Console
const FormattedLogger = require('moleculer').Loggers.Formatted
const {AsyncLocalStorage} = require('async_hooks')

const natsOptions = {
    type: 'NATS',
    options: {
        urls: ['nats://78.47.50.119'],
        user: 'console-service',
        pass: 'fdfdsdDW2ds21EW'
    }
}

const asyncLocalStorage = new AsyncLocalStorage()
asyncLocalStorage.enterWith(new Map())

class Logger extends ConsoleLogger {
    constructor(options, loggerBroker) {
        super(options)
        this.loggerBroker = loggerBroker
    }

    static build() {
        const broker = new ServiceBroker({
            namespace: 'sobix-console',
            transporter: natsOptions,
        })
        broker.start()

        return new Logger({}, broker)
    }

    getLogHandler(bindings) {
        const level = bindings ? this.getLogLevel(bindings.mod) : null
        if (!level) {
            return null
        }
        bindings.mod = bindings.mod ?? ''
        const levelIdx = FormattedLogger.LEVELS.indexOf(level)
        const formatter = this.getFormatter(bindings)

        return (type, args) => {
            const typeIdx = FormattedLogger.LEVELS.indexOf(type)
            if (typeIdx > levelIdx) {
                return
            }
            const store = asyncLocalStorage.getStore()
            if (store.get('traceId')) {
                try {
                    this.loggerBroker.broadcast('log.create', {
                        type: type,
                        serviceName: store.get('serviceName'),
                        serviceId: store.get('serviceId'),
                        traceId: store.get('traceId'),
                        time: Date.now(),
                        trace: args.reduce((r, v) => r += ` ${v}`, ''),
                    })
                } catch (e) {}
            }

            const pargs = formatter(type, args)
            switch(type) {
                case 'fatal':
                case 'error': return console.error(...pargs)
                case 'warn': return console.warn(...pargs)
                default: return console.log(...pargs)
            }
        }
    }
}

function handler(type, broker) {
    return (err) => {
        const store = asyncLocalStorage.getStore()
        return broker.broadcast('log.create', {
            type: type,
            serviceName: store.get('serviceName'),
            serviceId: store.get('serviceId'),
            traceId: store.get('traceId'),
            time: Date.now(),
            trace: err.stack,
            message: err.message
        })
    }
}

const logger = Logger.build()
const errorHandler = handler('ERROR', logger.loggerBroker)

module.exports = {
    middleware: {
        name: 'Logger',
        localAction(next, action) {
            return async (ctx) => {
                const store = new Map()
                    .set('serviceName', ctx.service.name)
                    .set('serviceId', ctx.nodeID)
                    .set('traceId', ctx.span.traceID)
                return asyncLocalStorage.run(store, () => next(ctx).catch((e) => {
                    try {
                        errorHandler(e)
                    } catch (e) {}
                    throw e
                }))
            }
        }
    },
    logger
}